# -*- coding: utf-8 -*-
import numpy as np
import pymoskito as pm

import matplotlib as mpl
import matplotlib.patches
import matplotlib.transforms


class MplPendulumVisualizer(pm.MplVisualizer):

    # parameters
    x_min_plot = -.85
    x_max_plot = .85
    y_min_plot = -.6
    y_max_plot = .6

    cart_height = .1
    cart_length = .2

    beam_height = .01
    beam_length = 1

    pendulum_shaft_height = 0.027
    pendulum_shaft_radius = 0.020

    pendulum_height = 0.5
    pendulum_radius = 0.005
    # end parameters

    def __init__(self, q_widget, q_layout):
        # setup canvas
        pm.MplVisualizer.__init__(self, q_widget, q_layout)
        self.axes.set_xlim(self.x_min_plot, self.x_max_plot)
        self.axes.set_ylim(self.y_min_plot, self.y_max_plot)
        self.axes.set_aspect("equal")

        # create patches
        self.beam = mpl.patches.Rectangle(xy=[-self.beam_length/2,
                                              -(self.beam_height
                                                + self.cart_height/2)],
                                          width=self.beam_length,
                                          height=self.beam_height,
                                          color="lightgrey")

        self.cart = mpl.patches.Rectangle(xy=[-self.cart_length/2,
                                              -self.cart_height/2],
                                          width=self.cart_length,
                                          height=self.cart_height,
                                          color="dimgrey")

        self.pendulum_shaft = mpl.patches.Circle(
            xy=[0, 0],
            radius=self.pendulum_shaft_radius,
            color="lightgrey",
            zorder=3)

        t = mpl.transforms.Affine2D().rotate_deg(180) + self.axes.transData
        self.pendulum = mpl.patches.Rectangle(
            xy=[-self.pendulum_radius, 0],
            width=2*self.pendulum_radius,
            height=self.pendulum_height,
            color=pm.colors.HKS07K100,
            zorder=2,
            transform=t)

        # add all to canvas
        self.axes.add_patch(self.beam)
        self.axes.add_patch(self.cart)
        self.axes.add_patch(self.pendulum_shaft)
        self.axes.add_patch(self.pendulum)

    # update callback
    def update_scene(self, x):
        cart_pos = x[0]
        phi = np.rad2deg(x[1])

        # cart and shaft
        self.cart.set_x(cart_pos - self.cart_length/2)
        self.pendulum_shaft.center = (cart_pos, 0)

        # pendulum
        ped_trafo = (mpl.transforms.Affine2D().rotate_deg(phi)
                     + mpl.transforms.Affine2D().translate(cart_pos, 0)
                     + self.axes.transData)
        self.pendulum.set_transform(ped_trafo)

        # update canvas
        self.canvas.draw()





# -*- coding: utf-8 -*-
import numpy as np
from scipy.linalg import solve_continuous_are
from collections import OrderedDict

import pymoskito as pm

from . import settings as st
from . import symbolic_calculation as symcalc


class LinearStateFeedback(pm.Controller):

    public_settings = OrderedDict([('poles', [-9, -7, -4, -2, -3, -2]),
                                   ("long pendulum", "u"),
                                   ("short pendulum", "o"),
                                   ('tick divider', 1)])

    def __init__(self, settings):
        # add specific private settings
        settings.update(input_order=0)
        settings.update(ouput_dim=1)
        settings.update(input_type='system_state')

        pm.Controller.__init__(self, settings)

        eq_state = calc_equilibrium(settings)
        # pole placement
        parameter = [st.m0, st.m1, st.m2, st.l1, st.l2, st.g, st.d0, st.d1, st.d2]
        self.A = symcalc.A_func(list(eq_state), parameter)
        self.B = symcalc.B_func(list(eq_state), parameter)
        self.C = symcalc.C
        self.K = pm.place_siso(self.A, self.B, self._settings['poles'])
        self.V = pm.calc_prefilter(self.A, self.B, self.C, self.K)
        # eig = np.linalg.eig(self.A - np.dot(self.B, self.K))

        self._logger.debug("Equilibrium: {}".format(eq_state.tolist()))
        self._logger.debug("Poles: {}".format(self._settings["poles"]))
        self._logger.debug("K: {}".format(self.K[0]))
        self._logger.debug("V: {}".format(self.V[0]))

    def _control(self, time, trajectory_values=None, feedforward_values=None,
                 input_values=None, **kwargs):
        # input abbreviations
        x = input_values
        yd = trajectory_values[0, 0]
        eq = kwargs.get("eq", None)

        if eq is None:
            eq = calc_closest_eq_state(self._settings, input_values)
        x = x - eq

        # this is a second version
        # x = calc_small_signal_state(self._settings, is_values)

        # u corresponds to a force [kg*m/s**2] = [N]
        u = - np.dot(self.K, x) + np.dot(self.V, yd)

        return u


class LinearStateFeedbackParLin(pm.Controller):

    public_settings = OrderedDict([('poles', [-9, -7, -4, -2, -3, -2]),
                                   ("long pendulum", "u"),
                                   ("short pendulum", "o"),
                                   ('tick divider', 1)])

    def __init__(self, settings):
        # add specific private settings
        settings.update(input_order=0)
        settings.update(ouput_dim=1)
        settings.update(input_type='system_state')

        pm.Controller.__init__(self, settings)

        eq_state = calc_equilibrium(settings)
        # pole placement
        parameter = [st.m0, st.m1, st.m2, st.l1, st.l2, st.g, st.d0, st.d1, st.d2]
        self.A = symcalc.A_func_par_lin(list(eq_state), parameter)
        self.B = symcalc.B_func_par_lin(list(eq_state), parameter)
        self.C = symcalc.C_par_lin
        self.K = pm.place_siso(self.A, self.B, self._settings['poles'])
        self.V = pm.calc_prefilter(self.A, self.B, self.C, self.K)
        # eig = np.linalg.eig(self.A - np.dot(self.B, self.K))

        self._logger.debug("Equilibrium: {}".format(eq_state.tolist()))
        self._logger.debug("Poles: {}".format(self._settings["poles"].tolist()))
        self._logger.debug("K: {}".format(self.K.tolist()[0]))
        self._logger.debug("V: {}".format(self.V[0]))

    def _control(self, time, trajectory_values=None, feedforward_values=None,
                 input_values=None, **kwargs):
        """

        :param time:
        :param trajectory_values:
        :param feedforward_values:
        :param input_values:
        :param kwargs:
        :return:
        """
        # input abbreviations
        x = input_values
        yd = trajectory_values[0, 0]
        eq = kwargs.get("eq", None)

        if eq is None:
            eq = calc_closest_eq_state(self._settings, input_values)
        x = x - eq

        # this is a second version
        # x = calc_small_signal_state(self._settings, is_values)

        # u corresponds to a acceleration [m/s**2]
        u = - np.dot(self.K, x) + np.dot(self.V, yd)

        return u


class LinearQuadraticRegulator(pm.Controller):

    public_settings = OrderedDict([("Q", [1, 0.063, 0.025, 0.025, 0.025, 0.025]),
                                   ("R", [1.0 / 144.0]),
                                   ("long pendulum", "u"),
                                   ("short pendulum", "o"),
                                   ("d0", 0.0),
                                   ('tick divider', 1)])

    def __init__(self, settings):
        # add specific private settings
        settings.update(input_order=0)
        settings.update(ouput_dim=1)
        settings.update(input_type='system_state')

        pm.Controller.__init__(self, settings)

        eq_state = calc_equilibrium(settings)
        # pole placement
        parameter = [st.m0, st.m1, st.m2, st.l1, st.l2, st.g, self._settings["d0"], st.d1, st.d2]
        self.A = symcalc.A_func(list(eq_state), parameter)
        self.B = symcalc.B_func(list(eq_state), parameter)
        self.C = symcalc.C

        # create Q and R matrix
        self.Q = np.diag(self._settings["Q"])
        self.R = np.diag(self._settings["R"])

        # try to solve linear riccati equation
        self.P = solve_continuous_are(self.A, self.B, self.Q, self.R)
        self.K = np.dot(np.dot(np.linalg.inv(self.R), self.B.T), self.P)
        self.V = pm.calc_prefilter(self.A, self.B, self.C, self.K)

        eig = np.linalg.eig(self.A - np.dot(self.B, self.K))

        self._logger.debug("equilibrium = " + str(eq_state))
        self._logger.debug("Q = " + str(self._settings["Q"]))
        self._logger.debug("R = " + str(self._settings["R"]))
        self._logger.debug("P = " + str(self.P))
        self._logger.debug("K = " + str(self.K))
        self._logger.debug("eig = " + str(eig))
        self._logger.debug("V = " + str(self.V[0][0]))

    def _control(self, time, trajectory_values=None, feedforward_values=None,
                 input_values=None, **kwargs):
        # input abbreviations
        x = input_values
        yd = trajectory_values[0, 0]
        eq = kwargs.get("eq", None)

        if eq is None:
            eq = calc_closest_eq_state(self._settings, x)
        x = x - eq

        # u corresponds to a force [kg*m/s**2] = [N]
        u = - np.dot(self.K, x) + np.dot(self.V, yd)

        return u


class LjapunovController(pm.Controller):

    public_settings = OrderedDict([
        ("k", 8),
        ("long pendulum", "u"),
        ("short pendulum", "o"),
        ("tick divider", 1)])

    def __init__(self, settings):
        settings.update(input_order=0)
        settings.update(output_order=1)
        settings.update(input_type="system_state")
        pm.Controller.__init__(self, settings)

        self.w = st.m1*st.l1/(st.m2*st.l2)

    def _control(self, time, trajectory_values=None, feedforward_values=None,
                 input_values=None, **kwargs):
        x1, x2, x3, x4, x5, x6 = input_values

        E0 = 0.5*st.m0*x2**2
        E1 = 0
        E2 = 0

        if self._settings["long pendulum"] == "u":
            E1 = 0.5*st.m1*st.l1**2*x4**2 + st.m1*st.g*st.l1*(np.cos(x3) + 1)
        elif self._settings["long pendulum"] == "o":
            E1 = 0.5*st.m1*st.l1**2*x4**2 + st.m1*st.g*st.l1*(np.cos(x3) - 1)

        if self._settings["short pendulum"] == "u":
            E2 = 0.5*st.m2*st.l2**2*x6**2 + st.m2*st.g*st.l2*(np.cos(x5) + 1)
        elif self._settings["short pendulum"] == "o":
            E2 = 0.5*st.m2*st.l2**2*x6**2 + st.m2*st.g*st.l2*(np.cos(x5) - 1)

        G = st.m0*x2*E0 + st.m1*st.l1*x4*np.cos(x3)*E1 + st.m2*st.l2*x6*np.cos(x5)*E2*self.w**2

        u_lja = -self._settings["k"]*G  # + (st.d1*E1*x4**2 + st.d2*E2*x6**2)/G

        # the u of the ljapunov method is a acceleration and the model require a force
        d0 = 0
        d1 = 0
        d2 = 0
        M1_star = 0
        M2_star = 0

        M = st.m0 + st.m1*(np.sin(x3))**2 + st.m2*(np.sin(x5))**2
        F1 = st.m1*np.sin(x3)*(st.g*np.cos(x3) - st.l1*x4**2)
        F2 = st.m2*np.sin(x5)*(st.g*np.cos(x5) - st.l2*x6**2)
        u = - F1 - F2 + u_lja*M + d0*x2 - (M1_star - d1*x4)*np.cos(x3)/st.l1 - (M2_star - d2*x6)*np.cos(x5)/st.l2

        return u


class SwingUpController(pm.Controller):
    """
    This class realise the swing up for equilibria with a arbitrary
    amount of turns of the pendulums

    The swing up of both pendulums require a adjustment of the
    integrator settings to
    rTol: 1e-09
    aTol: 1e-11
    """

    public_settings = OrderedDict([
        ("k", 10),
        ("long pendulum", "u"),
        ("short pendulum", "o"),
        ("poles", [-9, -7, -4, -2, -3, -2]),
        ("tick divider", 1)
    ])

    def __init__(self, settings):
        settings.update(input_order=0)
        settings.update(output_order=1)
        settings.update(input_type="system_state")
        self.module_settings = {"modules": settings["modules"]}
        pm.Controller.__init__(self, settings)

        # add dict with module information, because it was deleted in the base class
        settings.update(self.module_settings)
        settings["type"] = "LjapunovController"
        self.ljapunov = LjapunovController(settings)

        settings.update({"equilibrium": calc_equilibrium(settings)})
        settings.update(self.module_settings)
        settings["type"] = "LinearStateFeedback"
        self.linear_state_feedback = LinearStateFeedback(settings)

        self.switch = False

    def _control(self, time, trajectory_values=None, feedforward_values=None, input_values=None, **kwargs):
        x1, x2, x3, x4, x5, x6 = input_values

        # looking for the closest equilibrium
        eq_state = calc_closest_eq_state(self._settings, input_values)

        # we have to check several conditions
        #          phi1                  phi1_d
        a = (-0.2 + eq_state[2] <= x3 <= 0 + eq_state[2]) and (-0.2 <= x4 <= 0.5)
        #           phi2                 phi2_d
        b = (-0.2 + eq_state[4] <= x5 <= 0 + eq_state[4]) and (-0.2 <= x6 <= 0.5)
        #        phi1                  phi1_d
        c = (0 + eq_state[2] <= x3 <= 0.2 + eq_state[2]) and (-0.5 <= x4 <= 0.2)
        #          phi2                   phi2_d
        d = (0 + eq_state[4] <= x5 <= 0.2 + eq_state[4]) and (-0.5 <= x6 <= 0.2)

        eq_oo = False
        eq_ou = False
        eq_uo = False

        if self._settings["long pendulum"] == "o":
            eq_ou = True
        if self._settings["short pendulum"] == "o":
            eq_uo = True

        if eq_ou and eq_uo:
            eq_oo = True
            eq_ou = False
            eq_uo = False

        if (eq_oo and a and b)or (eq_oo and c and d) or (eq_ou and a) or (eq_ou and c) or (eq_uo and b) or (eq_uo and d):
            self.switch = True
        else:
            # this adjustment is only gut feeling
            e = (-0.3 + eq_state[2] <= x3 <= 0.3 + eq_state[2])
            f = (-0.3 + eq_state[4] <= x5 <= 0.3 + eq_state[4])
            if not (e and f):
                self.switch = False

        # print "t: ", t, "Switch: ", self.switch

        if self.switch:
            u = self.linear_state_feedback._control(time, trajectory_values,
                                                    input_values=input_values,
                                                    eq=eq_state)
        else:
            u = self.ljapunov._control(time,
                                       trajectory_values,
                                       input_values=input_values)

        return u.squeeze()


def calc_equilibrium(settings):
    equilibrium = np.zeros(6)
    if settings["long pendulum"] == "u":
        equilibrium[2] = np.pi
    if settings["short pendulum"] == "u":
        equilibrium[4] = np.pi

    return equilibrium


def calc_closest_eq_state(settings, state):
    # looking for the closest equilibrium for the two pendulums
    phi1 = float(state[2])
    phi2 = float(state[4])

    # consider the multiple pendulum states, because of the continuously angle
    # pendulum is on top: phi_eq = 2*PI*k
    # pendulum is at the bottom: phi_eq = (2*k + 1)*PI

    # pendulum 1 (long) is on top
    k1o = int(phi1/(2*np.pi))
    if (abs(phi1) % (2*np.pi)) > np.pi:
        k1o = k1o + np.sign(phi1)

    # pendulum 1 (long) is at the bottom
    k1u = int((phi1-np.pi)/(2*np.pi))
    if (abs(phi1 - np.pi) % (2*np.pi)) > np.pi:
        k1u = k1u + np.sign(phi1)

    # pendulum 2 (short) is on top
    k2o = int(phi2/(2*np.pi))
    if (abs(phi2) % (2*np.pi)) > np.pi:
        k2o = k2o + np.sign(phi2)

    # pendulum 2 (short) is at the bottom
    k2u = int((phi2 - np.pi)/(2*np.pi))
    if (abs(phi2 - np.pi) % (2*np.pi)) > np.pi:
        k2u = k2u + np.sign(phi2)

    eq_state = np.zeros(6)
    if settings["long pendulum"] == "o":
        eq_state[2] = 2*np.pi*k1o
    if settings["long pendulum"] == "u":
        eq_state[2] = (2*k1u + 1)*np.pi
    if settings["short pendulum"] == "o":
        eq_state[4] = 2*np.pi*k2o
    if settings["short pendulum"] == "u":
        eq_state[4] = (2*k2u + 1)*np.pi

    return eq_state


def calc_small_signal_state(settings, state):

    phi1 = (abs(state[2][0]) % (2*np.pi))*np.sign(state[2][0])
    phi2 = (abs(state[4][0]) % (2*np.pi))*np.sign(state[4][0])

    # transform the angle of the pendulums into an interval from -pi to +pi
    # pendulum 1 (long) is on top
    phi1_o = phi1
    if abs(phi1_o) > np.pi:
        phi1_o -= np.sign(state[2][0])*2*np.pi

    # pendulum 2 (short) is on top
    phi2_o = phi2
    if abs(phi2_o) > np.pi:
        phi2_o -= np.sign(state[4][0])*2*np.pi

    # transform the angle of the pendulums into an interval from 0 to 2pi
    # pendulum 1 (long) is at the bottom
    phi1_u = phi1
    if phi1_u < 0:
        phi1_u -= np.sign(state[2][0])*2*np.pi

    # pendulum 2 (short) is at the bottom
    phi2_u = phi2
    if phi2_u < 0:
        phi2_u -= np.sign(state[4][0])*2*np.pi

    # calc small signal state
    small_signal_state = np.zeros((6, 1))
    for idx, val in enumerate(state):
        small_signal_state[idx, 0] = val[0]

    if settings["long pendulum"] == "o":
        small_signal_state[2][0] = phi1_o - 0
    if settings["long pendulum"] == "u":
        small_signal_state[2][0] = phi1_u - np.pi
    if settings["short pendulum"] == "o":
        small_signal_state[4][0] = phi2_o - 0
    if settings["short pendulum"] == "u":
        small_signal_state[4][0] = phi2_u - np.pi

    return small_signal_state


pm.register_simulation_module(pm.Controller, LinearStateFeedback)
pm.register_simulation_module(pm.Controller, LinearStateFeedbackParLin)
pm.register_simulation_module(pm.Controller, LinearQuadraticRegulator)
pm.register_simulation_module(pm.Controller, LjapunovController)
pm.register_simulation_module(pm.Controller, SwingUpController)



from collections import OrderedDict

import numpy as np

import pymoskito as pm
from . import settings as st


class TwoTankSystem(pm.Model):
    public_settings = OrderedDict([('initial state', st.initial_states),
                                   ("AT", st.AT),
                                   ("hT", st.hT),
                                   ("AS1", st.AS1),
                                   ("AS2", st.AS2),
                                   ("g", st.g),
                                   ("K", st.K)])

    def __init__(self, settings):  # Constructor
        # add specific "private" settings
        settings.update(state_count=2)
        settings.update(input_count=1)
        settings.update({"output_info": {
            0: {"Name": "Level Tank1", "Unit": "m"},
            1: {"Name": "Level Tank2", "Unit": "m"},
        }})
        pm.Model.__init__(self, settings)

    def state_function(self, t, x, args):
        """
        Calculations of system state changes
        :param x: state
        :param t: time
        :type args: system input u
        """
        x1 = x[0]
        x2 = x[1]
        uA = args[0].squeeze()

        if x1 < 0:
            x1 = 0
        if x2 < 0:
            x2 = 0

        g = self.settings['g']
        K = self.settings['K']
        AS1 = self.settings['AS1']
        AS2 = self.settings['AS2']
        AT = self.settings['AT']

        dx1 = - AS1 / AT * np.sign(x1 - x2) * np.sqrt(2 * g * np.abs(x1 - x2)) + K / AT * uA
        dx2 = AS1 / AT * np.sign(x1 - x2) * np.sqrt(2 * g * np.abs(x1 - x2)) - AS2 / AT * np.sqrt(2 * g * x2)

        return np.array([dx1, dx2])

    def root_function(self, x):
        """
        in this case this means zero crossing detection for the water height elevation.
        :param x: state
        """
        x0 = x
        flag = False

        if x[0] <= 0:
            x0[0] = 0
            flag = True

        if x[1] <= 0:
            x0[1] = 0
            flag = True

        return flag, x0

    def calc_output(self, input_vector):
        return [input_vector[0], input_vector[1] + np.random.random(1)[0] / 100]


pm.register_simulation_module(pm.Model, TwoTankSystem)



# -*- coding: utf-8 -*-
import numpy as np
# -----------------------------------------------------------------------------
# global default settings for physical simulation
# -----------------------------------------------------------------------------

# -----------------------------------------------------------------------------
# parameters of the ball in tube system according to the
# real test station on the institute of control theory at the
# TU-Dresden

initial_state = [0, 0, 0, 0]

T = 0.24757155405           # s - time constant of the motor
d = 0.733277219979          # - damping ratio of the motor
k_s = 10.0291812304         # 1/Vs - gain motor
Vcc = 12                    # V - voltage supply

k_V = 6.28192637773e-05     # m**3 - factor between fan speed and airflow
k_L = 2.27269527471e-04     # N*s**2/m - factor between flow velocity in
#  the air gap squared and the buoyancy force
d_R = 64.4e-3               # m - diameter of the tube
d_R_out = 70.0e-3           # m - external diameter of the tube
tube_length = 1.5           # m -  length of the tube
d_B = 60.0e-3               # m - diameter of the ball
m = 2.8e-3                  # kg - mass of the ball
g = 9.81                    # m/s^2 - gravity

K = 10000                   # N/m - stiffness of the spring
D = 1                       # Ns/m - damping ratio of the spring

A_B = np.pi*d_B**2/4        # m**2 - cross sectional area of the ball
A_R = np.pi*d_R**2/4        # m**2 - cross sectional area of the tube
A_Sp = A_R - A_B            # m**2 - cross sectional area of air gap

scale = 2
visTubeLength = tube_length/scale
visTubeDiameter = d_R*scale
visBallDiameter = d_B*scale

# equilibrium
x1_e = A_Sp*np.sqrt(m*g/k_L)/k_V
x2_e = 0
x3_e = 0  # random
x4_e = 0
u_e = A_Sp*np.sqrt(m*g/k_L)/(k_V*k_s)

# -----------------------------------------------------------------------------
# export settings latex-style
# -----------------------------------------------------------------------------
latex_font_size = 14
label_size = 1*latex_font_size
title_size = 1.5*latex_font_size



#!/usr/bin/python2
# -*- coding: utf-8 -*-

import numpy as np
import os

observerList = ['LuenbergerObserver', 'HighGainObserver']
               
number = 1

#poles = [pol, pol, pol, pol]
pRange = [50, 2000]
pStepSize = 50
observer = observerList[number]

lines = ''



# load head file
filePath = os.path.join('A2_head.sray')
with open(filePath, 'r') as f:
    head = f.read()

pList= np.arange(pRange[0],pRange[1] + pStepSize, pStepSize)
lines += '\n'

for pol in pList:
    poles = [-pol, -pol, -pol, -pol]
    lines += '- name: C_' + observer + '_pol' + str(pol) + '\n'
    lines += '  clear previous: !!python/bool False' + '\n'
    lines += '\n'
    lines += '  solver:' + '\n'
    lines += '   type: VODESolver' + '\n'
    lines += '   initial state: [-4, 4.91768866, -0.2, -0.7]' + '\n'
    lines += '  controller: ' + '\n'
    lines += '   type: None\n'
    lines += '   poles: ' + str(poles) + '\n'
    lines += '\n'
    lines += '  trajectory:' +'\n'
    lines += '   type: HarmonicTrajectory' + '\n'
    lines += '   Amplitude: 1' + '\n'
    lines += '   Frequency: 1' + '\n'
    lines += '\n'
    lines += '  observer:' + '\n'    
    lines += '   type: '+ observer +'\n'
    lines += '   tick divider: 1' +'\n'
    lines += '   poles: ' + str(poles)+'\n'
    lines += '   initial state: [0, 0, 0, 0]'+'\n'
    if observer == 'LuenbergerObserver':
        lines += '   lin state: [0, 0, 0, 0]' +'\n'
    lines += '\n'                                        
            
#write results
#print os.path.pardir
dirPath = os.path.join(os.path.pardir, os.path.pardir, 'regimes', 'generated')
if not os.path.isdir(dirPath):
    os.makedirs(dirPath)

fileName = 'C2_' + observer +'_pol' + '_pRange(' + str(pRange[0])\
            + ',' + str(pRange[1]) + ')'+'.sreg'
            
filePath = os.path.join(dirPath, fileName)   

with open(filePath, 'w') as f:
    f.write(head)    
    f.write(lines)
    
    
    
    



# -*- coding: utf-8 -*-
import pymoskito as pm

# import custom modules
import model
import controller
import visualizer_mpl
# import visualizer_vtk


if __name__ == '__main__':
    # register model
    pm.register_simulation_module(pm.Model, model.PendulumModel)

    # register controller
    pm.register_simulation_module(pm.Controller, controller.BasicController)

    # register visualizer
    pm.register_visualizer(visualizer_mpl.MplPendulumVisualizer)
    # pm.register_visualizer(visualizer_vtk.VtkPendulumVisualizer)

    # start the program
    pm.run()



# -*- coding: utf-8 -*-

import numpy as np
from collections import OrderedDict

import pymoskito as pm


class OpenLoop(pm.Controller):
    """
    manual pwm input
    """
    public_settings = OrderedDict([("pwm", 160),
                                   ("tick divider", 1)])

    def __init__(self, settings):
        # add specific "private" settings
        settings.update(input_order=0)
        settings.update(output_dim=1)
        settings.update(input_type="system_state")

        pm.Controller.__init__(self, settings)

    def _control(self, time, trajectory_values=None, feedforward_values=None,
                 input_values=None, **kwargs):

        u = self._settings["pwm"]
        return np.array(u)

pm.register_simulation_module(pm.Controller, OpenLoop)



"""
Simple script that runs a pymoksito example

The name of the example must be given as first argument
"""

import sys
import os
from importlib import import_module
from PyQt5.QtWidgets import QApplication

import pymoskito as pm

if __name__ == '__main__':
    name = sys.argv[1]
    import_module("pymoskito.examples." + name)
    orig_dir = os.path.dirname(__file__)
    ex_dir = os.sep.join([orig_dir, os.pardir, "examples", name])
    res_dir = os.sep.join([orig_dir, "results", name])
    os.makedirs(res_dir, exist_ok=True)
    os.chdir(os.path.abspath(ex_dir))

    app = QApplication([])
    gui = pm.SimulationGui()
    gui.load_regimes_from_file("default.sreg")
    gui.actExitOnBatchCompletion.setChecked(True)
    gui._settings.setValue("path/simulation_results", os.path.abspath(res_dir))
    gui.start_regime_execution()

    os.chdir(orig_dir)
    app.exec_()



#!/usr/bin/python2
# -*- coding: utf-8 -*-

import numpy as np
import os
import settings as st

controllerList = ['FController', 'GController', 'JController',\
                'LSSController', 'PIFeedbackController']
               
#TODO: überlege, in welchem Punkt Regler linearisiert werden und dazu Pole
#####################################################
# Einstellungen
number = 2

aRange = [0.1, 3]
aStepSize = 0.1
freq = 0.1

#####################################################

lines = ''
controller = controllerList[number]


pol = st.poles[controllerList[number]]
print(controller)
print('pol:', pol)
if controller == 'PIFeedbackController':
    poles = [pol, pol, pol, pol, pol]
else:
    poles = [pol, pol, pol, pol]

# load head file
filePath = os.path.join('A2_head.sray')
with open(filePath, 'r') as f:
    head = f.read()

As = np.arange(aRange[0],aRange[1] + aStepSize, aStepSize)
lines += '\n'

for a in As:
        lines += '- name: A2_' + controller + '_f' + str(freq) + '_A'+str(a)+ '\n'
        lines += '  clear previous: !!python/bool False' + '\n'
        lines += '\n'
        lines += '  controller: ' + '\n'
        lines += '   type: '  + controller + '\n'
        lines += '   poles: ' + str(poles) + '\n'
        lines += '\n'
        lines += '  trajectory:' +'\n'
        lines += '   type: HarmonicTrajectory' + '\n'
        lines += '   Amplitude: ' + str(a) + '\n'
        lines += '   Frequency: ' + str(freq) + '\n'
        lines += '\n'
               
#write results
#print os.path.pardir
dirPath = os.path.join(os.path.pardir, os.path.pardir, 'regimes', 'generated')
if not os.path.isdir(dirPath):
    os.makedirs(dirPath)

fileName = 'A2_' + controller +'_f'+str(freq)+ '_aRange(' + str(aRange[0])\
            + ',' + str(aRange[1]) + ')'+'.sreg'
            
filePath = os.path.join(dirPath, fileName)   

with open(filePath, 'w') as f:
    f.write(head)    
    f.write(lines)
    
    
    
    



# -*- coding: utf-8 -*-
"""
Tools, functions and other funny things
"""
import copy
import logging
import os
import re

import matplotlib.gridspec as gridspec
import matplotlib.pyplot as plt
import pandas as pd
from PyQt5.QtGui import QColor
from PyQt5.QtWidgets import QPushButton

import numpy as np

logger = logging.getLogger(__name__)

__all__ = ["rotation_matrix_xyz", "get_resource", "sort_tree"]


def sort_lists(a, b):
    b = [x for (y, x) in sorted(zip(a, b))]
    a = sorted(a)
    return a, b


def get_resource(res_name, res_type="icons"):
    """
    Build absolute path to specified resource within the package

    Args:
        res_name (str): name of the resource
        res_type (str): subdir

    Return:
        str: path to resource
    """
    own_path = os.path.dirname(__file__)
    resource_path = os.path.abspath(os.path.join(own_path,
                                                 "resources",
                                                 res_type))
    return os.path.join(resource_path, res_name)


def sort_tree(data_list, sort_key_path):
    """
    Helper method for data sorting.

    Takes a list of simulation results and sorts them into a tree whose index is
    given by the sort_key_path.

    Args:
        data_list(list): List of simulation results
        sort_key_path(list): List of dictionary keys to sort for.

    Return:
        dict: sorted dictionary
    """
    result = {}
    for elem in data_list:
        temp_element = copy.deepcopy(elem)
        sort_name = get_sub_value(temp_element, sort_key_path)
        if sort_name not in result:
            result.update({sort_name: {}})

        while temp_element:
            val, keys = _remove_deepest(temp_element)
            if keys:
                _add_sub_value(result[sort_name], keys, val)

    return result


def get_sub_value(source, key_path):
    sub_dict = source
    for key in key_path:
        sub_dict = sub_dict[key]

    return sub_dict


def _remove_deepest(top_dict, keys=None):
    """
    Iterates recursively over dict and removes deepest entry.

    Args:
        top_dict (dict): dictionary
        keys (list): select entries to remove

    Return:
        tuple: entry and path to entry
    """
    if not keys:
        keys = []

    for key in list(top_dict.keys()):
        val = top_dict[key]
        if isinstance(val, dict):
            if val:
                keys.append(key)
                return _remove_deepest(val, keys)
            else:
                del top_dict[key]
                continue
        else:
            del top_dict[key]
            keys.append(key)
            return val, keys

    return None, None


def _add_sub_value(top_dict, keys, val):
    if len(keys) == 1:
        # we are here
        if keys[0] in top_dict:
            top_dict[keys[0]].append(val)
        else:
            top_dict.update({keys[0]: [val]})
        return

    # keep iterating
    if keys[0] not in top_dict:
        top_dict.update({keys[0]: {}})

    _add_sub_value(top_dict[keys[0]], keys[1:], val)
    return


def rotation_matrix_xyz(axis, angle, angle_dim):
    """
    Compute the rotation matrix for a rotation around a given axis with the
    angle :math:`\\varphi`.

    Args:
        axis (str): choose rotation axis "x", "y" or "z"
        angle (int or float): rotation angle :math:`\\varphi`
        angle_dim (str): choose "deg" for degree or "rad" for radiant

    Return:
        :obj:`numpy.ndarray`: rotation matrix
    """
    assert angle_dim == "deg" or angle_dim == "rad"
    assert axis == "x" or axis == "y" or axis == "z"
    x = 0
    y = 0
    z = 0

    if angle_dim == "deg":
        a = np.deg2rad(angle)
    else:
        a = angle

    if axis == "x":
        x = 1
        y = 0
        z = 0
    if axis == "y":
        x = 0
        y = 1
        z = 0
    if axis == "z":
        x = 0
        y = 0
        z = 1

    s = np.sin(a)
    c = np.cos(a)
    rotation_matrix = np.array([
        [c + x ** 2 * (1 - c),
         x * y * (1 - c) - z * s,
         x * z * (1 - c) + y * s],
        [y * x * (1 - c) + z * s,
         c + y ** 2 * (1 - c),
         y * z * (1 - c) - x * s],
        [z * x * (1 - c) - y * s,
         z * y * (1 - c) + x * s,
         c + z ** 2 * (1 - c)]])
    return rotation_matrix


class PlainTextLogger(logging.Handler):
    """
    Logging handler hat formats log data for line display
    """

    def __init__(self, settings, level=logging.NOTSET):
        logging.Handler.__init__(self, level)
        self.name = "PlainTextLogger"
        self.settings = settings

        formatter = logging.Formatter(
            fmt="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
            datefmt="%H:%M:%S")
        self.setFormatter(formatter)

        log_filter = PostFilter(invert=True)
        self.addFilter(log_filter)

        self.cb = None

    def set_target_cb(self, cb):
        self.cb = cb

    def emit(self, record):
        msg = self.format(record)
        if self.cb:
            clr = QColor(self.settings.value("log_colors/" + record.levelname,
                                             "#000000"))
            self.cb.setTextColor(clr)
            self.cb.append(msg)
        else:
            logging.getLogger().error("No callback configured!")


class PostFilter(logging.Filter):
    """
    Filter to sort out all not PostProcessing related log information
    """

    def __init__(self, invert=False):
        logging.Filter.__init__(self)
        self._invert = invert
        self.exp = re.compile(r"Post|Meta|Process")

    def filter(self, record):
        m = self.exp.match(record.name)
        if self._invert:
            return not bool(m)
        else:
            return bool(m)


def swap_cols(arr, frm, to):
    """ Swap the column `frm` from a given index `to` the given index.
    """
    arr[:, [frm, to]] = arr[:, [to, frm]]
    return arr


def swap_rows(arr, frm, to):
    """ Swap the rows `frm` from a given index `to` the given index.
    """
    if len(arr.shape) == 1:
        arr[[frm, to]] = arr[[to, frm]]
    elif len(arr.shape) == 2:
        arr[[frm, to], :] = arr[[to, frm], :]
    return arr


def get_figure_size(scale):
    """
    calculate optimal figure size with the golden ratio
    :param scale:
    :return:
    """
    # TODO: Get this from LaTeX using \the\textwidth
    fig_width_pt = 448.13095
    inches_per_pt = 1.0 / 72.27  # Convert pt to inch (stupid imperial system)
    golden_ratio = (np.sqrt(5.0) - 1.0) / 2.0  # Aesthetic ratio
    fig_width = fig_width_pt * inches_per_pt * scale  # width in inches
    fig_height = fig_width * golden_ratio  # height in inches
    fig_size = [fig_width, fig_height]
    return fig_size


class Exporter:
    def __init__(self, **kwargs):
        data_points = kwargs.get('data_points', None)

        if data_points is None:
            raise Exception("Given data points are None!")

        # build pandas data frame
        self.df = pd.DataFrame.from_dict(data_points)

        if 'time' in self.df.columns:
            self.df.set_index('time', inplace=True)

    def export_png(self, file_name):
        fig = plt.figure(figsize=(10, 6))
        gs = gridspec.GridSpec(1, 1, hspace=0.1)
        axes = plt.Subplot(fig, gs[0])

        for col in self.df.columns:
            self.df[col].plot(ax=axes, label=col)

        axes.legend(bbox_to_anchor=(0., 1.02, 1., .102), loc=4,
                    ncol=4, mode="expand", borderaxespad=0., framealpha=0.5)

        axes.grid(True)
        if self.df.index.name == 'time':
            axes.set_xlabel(r"Time (s)")

        fig.add_subplot(axes)

        fig.savefig(file_name, dpi=300)

    def export_csv(self, file_name, sep=','):
        self.df.to_csv(file_name, sep=sep)


def create_button_from_action(action):
    """
    QPushButton that is generated from a QAction
    """
    btn = QPushButton()
    btn.setIcon(action.icon())
    btn.setToolTip(action.toolTip())
    btn.clicked.connect(action.activate)
    return btn


# -*- coding: utf-8 -*-

import os
from collections import OrderedDict

import numpy as np

import pymoskito as pm
from . import settings as st


class CppHighGainObserver(pm.CppBase, pm.Observer):
    """
    High Gain Observer implemented in cpp with pybind11
    """
    public_settings = OrderedDict([
        ("initial state", st.initial_states),
        ("poles", st.poles_obs),
        ("AT", st.AT),
        ("hT", st.hT),
        ("AS1", st.AS1),
        ("AS2", st.AS2),
        ("g", st.g),
        ("K", st.K),
        ("dt [s]", 0.1),
        ("tick divider", 1),
    ])

    def __init__(self, settings):
        settings.update(output_dim=2)

        m_path = os.path.dirname(__file__)
        if 'tanksystem' not in m_path:
            m_path += '/pymoskito/examples/tanksystem/src/'
        else:
            m_path += '/src/'

        addLib = "add_library(pybind11 INTERFACE)\n"
        addLib += "target_include_directories(pybind11 INTERFACE $(VENV)/lib/$(PYVERS)/site-packages/pybind11/include/)\n"

        pm.Observer.__init__(self, settings)
        pm.CppBase.__init__(self,
                            module_path=m_path,
                            module_name='Observer',
                            binding_class_name="binding_Observer",
                            # If pybind is not global installed, use the uncomment the following line and set $(VENV) 
                            # as also as $(PYVERS) in the target_include_directories string.
                            # additional_lib={'pybind11': addLib}
                            )

        self.obs = self.get_class_from_module().HighGainObserver(self._settings["AT"],
                                                                 self._settings["AS1"],
                                                                 self._settings["AS2"],
                                                                 self._settings["K"],
                                                                 self._settings['dt [s]'])
        self.obs.set_initial_state(np.array(self._settings["initial state"]))
        self.obs.set_gain(np.array(self._settings["poles"]))

    def _observe(self, time, system_input, system_output):
        if system_input is None:
            return np.array(self._settings["initial state"])

        y = system_output[1]
        uA = system_input[0]
        res = self.obs.compute(y, uA)

        return np.array(res)


pm.register_simulation_module(pm.Observer, CppHighGainObserver)



# -*- coding: utf-8 -*-
from collections import OrderedDict
import numpy as np

import pymoskito as pm

from . import settings as st


class TwoPendulumModel(pm.Model):
    """
    Implementation of the two pendulum on a cart system
    this is implemented as a point mass model
    """
    public_settings = OrderedDict([("initial state", st.initial_state),
                                   ("m0*", st.m0_star),
                                   ("d0", st.d0),
                                   ("m1*", st.m1_star),
                                   ("l1*", st.l1_star),
                                   ("J_DP1", st.J_DP1),
                                   ("d1", st.d1),
                                   ("m2*", st.m2_star),
                                   ("l2*", st.l2_star),
                                   ("J_DP2", st.J_DP2),
                                   ("d2", st.d2),
                                   ("g", st.g),
                                   ])

    def __init__(self, settings):
        # conversion from degree to radiant
        settings["initial state"][2:] = [
            val*np.pi/180.0 for val in settings["initial state"][2:]]

        # add specific "private" settings
        settings.update(state_count=6)
        settings.update(input_count=1)
        settings.update({"output_info": {
            0: {"Name": "cart-position", "Unit": "m"},
            1: {"Name": "cart-velocity", "Unit": "m"},
            2: {"Name": "long-pendulum-angle", "Unit": "rad/s"},
            3: {"Name": "long-pendulum-angular-velocity", "Unit": "rad/s"},
            4: {"Name": "short-pendulum-angle", "Unit": "rad"},
            5: {"Name": "short-pendulum-angular-velocity", "Unit": "rad/s"},
        }})
        pm.Model.__init__(self, settings)

        # shortcuts for readability
        self.d0 = self._settings["d0"]
        self.d1 = self._settings["d1"]
        self.d2 = self._settings["d2"]
        self.g = self._settings["g"]

        self.l1 = self._settings["J_DP1"]/(self._settings["m1*"]*self._settings["l1*"])
        self.l2 = self._settings["J_DP2"]/(self._settings["m2*"]*self._settings["l2*"])

        self.m1 = (self._settings["m1*"]*self._settings["l1*"])**2/self._settings["J_DP1"]
        self.m2 = (self._settings["m2*"]*self._settings["l2*"])**2/self._settings["J_DP2"]

        self.m0 = self._settings["m0*"] + (self._settings["m1*"] - self.m1) + (self._settings["m2*"] - self.m2)

    def state_function(self, t, x, args):
        """
        Calculations of system state changes
        :param x: state
        :param t: time
        :type args: system input force on the cart
        """

        # definitional
        x1 = x[0]
        x2 = x[1]
        x3 = x[2]
        x4 = x[3]
        x5 = x[4]
        x6 = x[5]
        F_star = args[0].squeeze()
        M1_star = 0
        M2_star = 0

        # transformation of the input
        M = self.m0 + self.m1*(np.sin(x3))**2 + self.m2*(np.sin(x5))**2
        F1 = self.m1*np.sin(x3)*(self.g*np.cos(x3) - self.l1*x4**2)
        F2 = self.m2*np.sin(x5)*(self.g*np.cos(x5) - self.l2*x6**2)
        u = (F1 +
             F2 +
             (F_star - self.d0*x2) +
             (M1_star - self.d1*x4)*np.cos(x3)/self.l1 +
             (M2_star - self.d2*x6)*np.cos(x5)/self.l2)/M

        dx1 = x2
        dx2 = u
        dx3 = x4
        dx4 = self.g*np.sin(x3)/self.l1 + u*np.cos(x3)/self.l1 + (M1_star - self.d1*x4)/(self.m1*self.l1**2)
        dx5 = x6
        dx6 = self.g*np.sin(x5)/self.l2 + u*np.cos(x5)/self.l2 + (M2_star - self.d2*x6)/(self.m2*self.l2**2)

        dx = np.array([dx1, dx2, dx3, dx4, dx5, dx6])
        return dx

    def root_function(self, x):
        return [False]

    def check_consistency(self, x):
        """
        Check something
        """
        pass

    def calc_output(self, input):
        """
        return cart position as output
        :param input: input values
        :return: cart position
        """
        return input[0]


class TwoPendulumRigidBodyModel(pm.Model):
    """
    Implementation of the two pendulum on a cart system
    this is implemented as rigid body model
    """
    public_settings = OrderedDict([("initial state", st.initial_state),
                                   ("m0*", st.m0_star),
                                   ("d0", st.d0),
                                   ("m1*", st.m1_star),
                                   ("l1*", st.l1_star),
                                   ("J_DP1", st.J_DP1),
                                   ("d1", st.d1),
                                   ("m2*", st.m2_star),
                                   ("l2*", st.l2_star),
                                   ("J_DP2", st.J_DP2),
                                   ("d2", st.d2),
                                   ("g", st.g),
                                   ])

    def __init__(self, settings):
        # conversion from degree to radiant
        settings["initial state"][2:] = settings["initial state"][2:]*np.pi/180
        # add specific "private" settings
        settings.update(state_count=6)
        settings.update(input_count=1)
        settings.update({"output_info": {
            0: {"Name": "cart-position", "Unit": "m"},
            1: {"Name": "cart-velocity", "Unit": "m"},
            2: {"Name": "long-pendulum-angle", "Unit": "rad/s"},
            3: {"Name": "long-pendulum-angular-velocity", "Unit": "rad/s"},
            4: {"Name": "short-pendulum-angle", "Unit": "rad"},
            5: {"Name": "short-pendulum-angular-velocity", "Unit": "rad/s"},
        }})
        pm.Model.__init__(self, settings)

        print(self._settings)
        # shortcuts for readability
        self.d0 = self._settings["d0"]
        self.d1 = self._settings["d1"]
        self.d2 = self._settings["d2"]
        self.g = self._settings["g"]

        self.l1_star = self._settings["l1*"]
        self.l2_star = self._settings["l2*"]

        self.J_DP1 = self._settings["J_DP1"]
        self.J_DP2 = self._settings["J_DP2"]

        self.m0_star = self._settings["m0*"]
        self.m1_star = self._settings["m1*"]
        self.m2_star = self._settings["m2*"]

    def state_function(self, t, x, args):
        """
        Calculations of system state changes
        :param x: state
        :param t: time
        :type args: system input force on the cart
        """

        # definitional
        x1 = x[0]
        x2 = x[1]
        x3 = x[2]
        x4 = x[3]
        x5 = x[4]
        x6 = x[5]
        F = args[0].squeeze()
        M1 = 0
        M2 = 0

        # transformation of the input
        term1 = (self.m0_star
                 + self.m1_star
                 + self.m2_star
                 - self.m1_star**2*self.l1_star**2*(np.cos(x3))**2/self.J_DP1
                 - self.m2_star**2*self.l2_star**2*(np.cos(x5))**2/self.J_DP2)
        term2 = self.m1_star*self.l1_star*(np.cos(x3))*(M1
                                                        - self.d1*x4
                                                        + self.m1_star*self.l1_star*self.g*np.sin(x3))/self.J_DP1
        term3 = self.m2_star*self.l2_star*(np.cos(x5))*(M2
                                                        - self.d2*x6
                                                        + self.m2_star*self.l2_star*self.g*np.sin(x5))/self.J_DP2
        term4 = (F
                 - self.d0*x2
                 - self.m1_star*self.l1_star*x4**2*np.sin(x3)
                 - self.m2_star*self.l2_star*x6**2*np.sin(x5))

        dx1 = x2
        dx2 = (term2 + term3 + term4)/term1
        dx3 = x4
        dx4 = (self.m1_star*self.l1_star*np.cos(x3)*dx2
               + M1
               - self.d1*x4
               + self.m1_star*self.l1_star*self.g*np.sin(x3))/self.J_DP1
        dx5 = x6
        dx6 = (self.m2_star*self.l2_star*np.cos(x5)*dx2
               + M2
               - self.d2*x6
               + self.m2_star*self.l2_star*self.g*np.sin(x5))/self.J_DP2

        dx = np.array([dx1, dx2, dx3, dx4, dx5, dx6])
        return dx

    def root_function(self, x):
        return [False]

    def check_consistency(self, x):
        """
        Check something
        """
        pass

    def calc_output(self, input):
        """
        return cart position as output
        :param input: input values
        :return: cart position
        """
        return input[0]


class TwoPendulumModelParLin(pm.Model):
    """
    Implementation of the two pendulum on a cart system, with a acceleration as input
    this is the partial linerarisation of the TwoPendulumModel
    """
    public_settings = OrderedDict([("initial state", st.initial_state),
                                   ("m0*", st.m0_star),
                                   ("d0", st.d0),
                                   ("m1*", st.m1_star),
                                   ("l1*", st.l1_star),
                                   ("J_DP1", st.J_DP1),
                                   ("d1", st.d1),
                                   ("m2*", st.m2_star),
                                   ("l2*", st.l2_star),
                                   ("J_DP2", st.J_DP2),
                                   ("d2", st.d2),
                                   ("g", st.g),
                                   ])

    def __init__(self, settings):
        # conversion from degree to radiant
        settings["initial state"][2:] = settings["initial state"][2:]*np.pi/180
        # add specific "private" settings
        settings.update(state_count=6)
        settings.update(input_count=1)
        settings.update({"output_info": {
            0: {"Name": "cart-position", "Unit": "m"},
            1: {"Name": "cart-velocity", "Unit": "m"},
            2: {"Name": "long-pendulum-angle", "Unit": "rad/s"},
            3: {"Name": "long-pendulum-angular-velocity", "Unit": "rad/s"},
            4: {"Name": "short-pendulum-angle", "Unit": "rad"},
            5: {"Name": "short-pendulum-angular-velocity", "Unit": "rad/s"},
        }})
        pm.Model.__init__(self, settings)

        # shortcuts for readability
        self.d0 = self._settings["d0"]
        self.d1 = self._settings["d1"]
        self.d2 = self._settings["d2"]
        self.g = self._settings["g"]

        self.l1 = self._settings["J_DP1"]/(self._settings["m1*"]*self._settings["l1*"])
        self.l2 = self._settings["J_DP2"]/(self._settings["m2*"]*self._settings["l2*"])

        self.m1 = (self._settings["m1*"]*self._settings["l1*"])**2/self._settings["J_DP1"]
        self.m2 = (self._settings["m2*"]*self._settings["l2*"])**2/self._settings["J_DP2"]

        self.m0 = self._settings["m0*"] + (self._settings["m1*"] - self.m1) + (self._settings["m2*"] - self.m2)

    def state_function(self, t, x, args):
        """
        Calculations of system state changes
        :param x: state
        :param t: time
        :type args: system input force on the cart
        """

        # definitional
        x1 = x[0]
        x2 = x[1]
        x3 = x[2]
        x4 = x[3]
        x5 = x[4]
        x6 = x[5]
        u = args[0].squeeze()

        dx1 = x2
        dx2 = u
        dx3 = x4
        dx4 = self.g*np.sin(x3)/self.l1 - (self.d1*x4)/(self.m1*self.l1**2) + np.cos(x3)*u/self.l1
        dx5 = x6
        dx6 = self.g*np.sin(x5)/self.l2 - (self.d2*x6)/(self.m2*self.l2**2) + np.cos(x5)*u/self.l2

        dx = np.array([dx1, dx2, dx3, dx4, dx5, dx6])
        return dx

    def root_function(self, x):
        return [False]

    def check_consistency(self, x):
        """
        Check something
        """
        pass

    def calc_output(self, input):
        """
        return cart position as output
        :param input: input values
        :return: cart position
        """
        return input[0]

pm.register_simulation_module(pm.Model, TwoPendulumModel)
pm.register_simulation_module(pm.Model, TwoPendulumModelParLin)
pm.register_simulation_module(pm.Model, TwoPendulumRigidBodyModel)



# -*- coding: utf-8 -*-
import numpy as np
from collections import OrderedDict

import pymoskito as pm


# class begin
class BasicController(pm.Controller):
    public_settings = OrderedDict([("poles", [-2, -2, -2, -2])
                                   ])

    # init
    def __init__(self, settings):
        settings.update(input_order=0)
        settings.update(input_type="system_state")

        pm.Controller.__init__(self, settings)

        # model parameters
        g = 9.81  # gravity [m/s^2]
        M = 4.2774  # cart mass [kg]
        D = 10  # cart friction constant [Ns/m]
        m = 0.3211  # pendulum mass [kg]
        d = 0.023  # pendulum friction constant [Nms]
        l = 0.3533  # pendulum length [m]
        J = 0.072  # pendulum moment of intertia [kg*m^2]

        # the system matrices after linearization in phi=PI
        z = (M + m) * J - (m * l) ** 2
        A = np.array([[0, 0, 1, 0],
                      [0, 0, 0, 1],
                      [0, (m * l) ** 2 * g / z, -J * D / z, m * l * d / z],
                      [0, -(M + m) * m * l * g / z, m * l * D / z,
                       -(M + m) * d / z]
                      ])
        B = np.array([[0],
                      [0],
                      [J / z],
                      [-l * m / z]
                      ])
        C = np.array([[1, 0, 0, 0]])

        # the equilibrium state as a vector
        self._eq_state = np.array([0, np.pi, 0, 0])

        # pole placement of linearized state feedback
        self._K = pm.controltools.place_siso(A, B, self._settings["poles"])
        self._V = pm.controltools.calc_prefilter(A, B, C, self._K)

    # control
    def _control(self, time, trajectory_values=None, feedforward_values=None,
                 input_values=None, **kwargs):
        x = input_values - self._eq_state
        yd = trajectory_values - self._eq_state[0]
        output = - np.dot(self._K, x) + np.dot(self._V, yd[0])

        return output



# -*- coding: utf-8 -*-
import sys
import os
os.environ["QT_AUTO_SCREEN_SCALE_FACTOR"] = "1"  # use high dpi settings
import importlib
from PyQt5.QtWidgets import QApplication

import pymoskito as pm

if __name__ == '__main__':
    # This part lets us execute this example with:
    #  $python -m pymoskito.examples.name
    pkg_path = os.path.dirname(os.path.abspath(__file__))
    pkg_name = pkg_path.split(os.path.sep)[-1]
    parent_dir = os.path.dirname(pkg_path)
    sys.path.insert(0, parent_dir)

    if __package__ is None or __package__ == '':
        importlib.import_module(pkg_name)
        __package__ = pkg_name

    # create an Application instance (needed)
    app = QApplication([])
    prog = None

    if 1:
        # create simulator
        prog = pm.SimulationGui()

        # load default config
        prog.load_regimes_from_file(os.path.join(parent_dir,
                                                 pkg_name,
                                                 "default.sreg"))
        prog.apply_regime_by_name("test-nonlinear-f")
    else:
        prog = pm.PostProcessor()

    # show gui
    prog.show()

    app.exec_()



# -*- coding: utf-8 -*-
import numpy as np
from collections import OrderedDict

import pymoskito as pm
import pymoskito.controltools

from . import settings as st
from .linearization import linearise_system
#import end


class FController(pm.Controller):
    """
    Controller created by changing f(x)
    """
    public_settings = OrderedDict([("poles", [-3.1, -3.1, -3.1, -3.1]),
                                   ("source", "system_state"),
                                   ("tick divider", 1),
                                   ])

    def __init__(self, settings):
        # add specific "private" settings
        settings.update(input_order=4)
        settings.update(output_dim=1)
        settings.update(input_type=settings["source"])

        pm.Controller.__init__(self, settings)
        self._output = np.zeros((1, ))

        # run pole placement
        self.K = pymoskito.controltools.char_coefficients(self._settings["poles"])

    def _control(self, time, trajectory_values=None, feedforward_values=None,
                 input_values=None, **kwargs):
        # input abbreviations
        x1, x2, x3, x4 = input_values
        yd = trajectory_values[0]

        # calculate nonlinear terms phi
        phi1 = x1
        phi2 = x2
        phi3 = -st.B*st.G*np.sin(x3)
        phi4 = -st.B*st.G*x4*np.cos(x3)

        # calculate fictional input v
        v = (yd[4]
             + self.K[3] * (yd[3] - phi4)
             + self.K[2] * (yd[2] - phi3)
             + self.K[1] * (yd[1] - phi2)
             + self.K[0] * (yd[0] - phi1)
             )

        # calculate a(x)
        a = -st.B*st.G*np.cos(x3)
        # calculate b(x)
        b = st.B*st.G*x4**2*np.sin(x3)
        
        # calculate u
        u = (v-b)/a

        # transform back
        self._output = (u * (st.M*x1**2 + st.J + st.Jb)
                        + st.M*(2*x1*x2*x4 + st.G*x1*np.cos(x3)))
        
        return self._output


class GController(pm.Controller):
    """
    Controller created by changing g(x)
    """
    public_settings = OrderedDict([("poles", [-2.8, -2.8, -2.8, -2.8]),
                                   ("source", "system_state"),
                                   ("tick divider", 1),
                                   ])

    def __init__(self, settings):
        # add specific "private" settings
        settings.update(input_order=4)
        settings.update(output_dim=1)
        settings.update(input_type=settings["source"])

        pm.Controller.__init__(self, settings)
        self._output = np.zeros((1, ))

        # run pole placement
        self.K = pm.char_coefficients(self._settings["poles"])

    def _control(self, time, trajectory_values=None, feedforward_values=None,
                 input_values=None, **kwargs):
        # input abbreviations
        x = input_values
        yd = trajectory_values[0]

        # calculate nonlinear terms phi
        phi1 = x[0]
        phi2 = x[1]
        phi3 = -st.B*st.G*np.sin(x[2]) + st.B*x[0]*x[3]**2
        phi4 = -st.B*st.G*x[3]*np.cos(x[2]) + st.B*x[1]*x[3]**2

        # calculate fictional input v
        v = (yd[4]
             + self.K[3] * (yd[3] - phi4)
             + self.K[2] * (yd[2] - phi3)
             + self.K[1] * (yd[1] - phi2)
             + self.K[0] * (yd[0] - phi1)
             )

        # calculate a(x)
        a = -st.B*st.G*np.cos(x[2]) + 2*st.B*x[1]*x[3]

        # calculate b(x)
        b = st.B**2*x[0]*x[3]**4 + st.B*st.G*(1 - st.B)*x[3]**2*np.sin(x[2])

        # calculate u
        u = (v-b)/a
        return u


class JController(pm.Controller):
    """
    Controller based on the standard jacobian approximation.
    """
    public_settings = OrderedDict([("poles", [-2, -2, -2, -2]),
                                   ("source", "system_state"),
                                   ("tick divider", 1),
                                   ])

    def __init__(self, settings):
        # add specific "private" settings
        settings.update(input_order=4)
        settings.update(output_dim=1)
        settings.update(input_type=settings["source"])

        pm.Controller.__init__(self, settings)
        self._output = np.zeros((1, ))

        # run pole placement
        self.K = pm.char_coefficients(self._settings["poles"])

    def _control(self, time, trajectory_values=None, feedforward_values=None,
                 input_values=None, **kwargs):
        # input abbreviations
        x = input_values
        yd = trajectory_values[0]

        # calculate linear terms phi
        phi1 = x[0]
        phi2 = x[1]
        phi3 = -st.B*st.G*x[2]
        phi4 = -st.B*st.G*x[3]

        # calculate fictional input v
        v = (yd[4]
             + self.K[3] * (yd[3] - phi4)
             + self.K[2] * (yd[2] - phi3)
             + self.K[1] * (yd[1] - phi2)
             + self.K[0] * (yd[0] - phi1)
             )

        # calculate a(x)
        a = -st.B*st.G/(st.J + st.Jb)

        # calculate b(x)
        b = st.B*st.M*st.G**2*x[0]/(st.J + st.Jb)

        # calculate u
        u = (v-b)/a
        return u


#class begin
class LSSController(pm.Controller):
    r"""
    Linear state space controller
    
    This controller is based on the linearized system. The steady state is
    given by :math:`(\boldsymbol{x}^e, \tau^e)` .
    """
    public_settings = OrderedDict([("poles", [-3.1, -3.1, -3.1, -3.1]),
                                   ("source", "system_state"),
                                   ("steady state", [0, 0, 0, 0]),
                                   ("steady tau", 0),
                                   ("tick divider", 1)
                                   ])
    #init
    def __init__(self, settings):
        # add specific "private" settings
        settings.update(input_order=0)
        settings.update(output_dim=1)
        settings.update(input_type=settings["source"])

        pm.Controller.__init__(self, settings)
        self._output = np.zeros((1, ))

        # run pole placement
        a_mat, b_mat, c_mat = linearise_system(self._settings["steady state"],
                                               self._settings["steady tau"])
        self.K = pm.place_siso(a_mat, b_mat, self._settings["poles"])
        self.V = pm.calc_prefilter(a_mat, b_mat, c_mat, self.K)

    #control
    def _control(self, time, trajectory_values=None, feedforward_values=None,
                    input_values=None, **kwargs):
        # input abbreviations
        yd = trajectory_values[:, 0]

        self._output = -self.K @ input_values + yd * self.V
        return self._output
#class end


class PIXController(pm.Controller):
    r"""
    Linear statespace controller
    
    This controller is based on the linearized system, extended with an 
    integrator. The steady state is
    given by :math:`(\boldsymbol{x}^e, \tau^e)` .
    """
    public_settings = OrderedDict([("poles", [-3.1, -3.1, -3.1, -3.1]),
                                   ("source", "system_state"),
                                   ("steady state", [0, 0, 0, 0]),
                                   ("steady tau", 0),
                                   ("step width", .01),
                                   ('Ki', -4),
                                   ("I Limit", 0.1),
                                   ("tick divider", 1)
                                   ])

    def __init__(self, settings):
        # add specific "private" settings
        settings.update(input_order=0)
        settings.update(output_dim=1)
        settings.update(input_type=settings["source"])

        pm.Controller.__init__(self, settings)
        self._error = 0
        self._output = np.zeros((1, ))

        # run pole placement
        a_mat, b_mat, c_mat = linearise_system(self._settings["steady state"],
                                               self._settings["steady tau"])
        self.K = pm.place_siso(a_mat, b_mat, self._settings["poles"])
        self.V = pm.calc_prefilter(a_mat, b_mat, c_mat, self.K)
        self.h = self._settings["tick divider"] * self._settings["step width"]

    def _control(self, time, trajectory_values=None, feedforward_values=None,
                 input_values=None, **kwargs):
        # input abbreviations
        x = input_values
        yd = trajectory_values

        e = self._error + self.h * (yd[0] - x[0])

        # saturate error
        if not abs(e) > self._settings["I Limit"]:
            self._error = e

        # calculate u
        u = -self.K @ x + yd[0]*self.V + self._settings["Ki"] * self._error
        return u


#register
pm.register_simulation_module(pm.Controller, FController)
pm.register_simulation_module(pm.Controller, GController)
pm.register_simulation_module(pm.Controller, JController)
pm.register_simulation_module(pm.Controller, LSSController)
pm.register_simulation_module(pm.Controller, PIXController)



# -*- coding: utf-8 -*-
from collections import OrderedDict

import numpy as np
import sympy as sp

import pymoskito as pm

from . import settings as st

"""
Feedforward implementations for the ball and beam system
"""


class LinearFeedforward(pm.Feedforward):
    """
    A Feedforward that compensates the linear part of the system dynamics.
    """

    public_settings = OrderedDict([
        ("r0", 3),
        ("M", st.M),
        ("R", st.R),
        ("J", st.J),
        ("Jb", st.Jb),
        ("G", st.G),
        ("B", st.B),
        ("tick divider", 1)
    ])

    def __init__(self, settings):
        # add specific "private" settings
        settings.update(input_order=4)
        settings.update(output_dim=1)

        pm.Feedforward.__init__(self, settings)

    def _calc_theta_t(self, rdd):
        theta_t = -rdd / (self._settings["B"] * self._settings["G"])
        return theta_t

    def _feedforward(self, t, yd):
        theta = []

        # need the second derivative of theta
        for i in range(2 + 1):
            theta.append(self._calc_theta_t(yd[i + 2]))

        tau = ((self._settings['M'] * self._settings['r0'] ** 2
                + self._settings['J']
                + self._settings['Jb']) * theta[2]
               + self._settings['M'] * self._settings['G'] * yd[0])

        return tau


pm.register_simulation_module(pm.Feedforward, LinearFeedforward)



# -*- coding: utf-8 -*-

import os
from collections import OrderedDict

import numpy as np
import sympy as sp
from sympy.solvers import solve

import pymoskito as pm
from . import settings as st


class OpenLoop(pm.Controller):
    """
    manual pwm input
    """
    public_settings = OrderedDict([("uA", 10),
                                   ("tick divider", 1)])

    def __init__(self, settings):
        # add specific "private" settings
        settings.update(input_order=0)
        settings.update(output_dim=1)
        settings.update(input_type="system_state")

        pm.Controller.__init__(self, settings)

    def _control(self, time, trajectory_values=None, feedforward_values=None,
                 input_values=None, **kwargs):
        u = self._settings["uA"]
        return np.array(u)


class CppPIDController(pm.CppBase, pm.Controller):
    """
    PID Controller implemented in cpp with pybind11
    """
    public_settings = OrderedDict([
        ("Kp", st.Kp),
        ("Ti", st.Ti),
        ("Td", st.Td),
        ("dt [s]", 0.1),
        ("output_limits", st.limits_ctrl),
        ("input_state", st.input_ctrl),
        ("tick divider", 1),
    ])

    def __init__(self, settings):
        # add specific "private" settings
        settings.update(input_order=0)
        settings.update(output_dim=1)
        settings.update(input_type="system_state")

        m_path = os.path.dirname(__file__)
        if 'tanksystem' not in m_path:
            m_path += '/pymoskito/examples/tanksystem/src/'
        else:
            m_path += '/src/'

        addLib = "add_library(pybind11 INTERFACE)\n"
        addLib += "target_include_directories(pybind11 INTERFACE $(VENV)/lib/$(PYVERS)/site-packages/pybind11/include/)\n"

        pm.Controller.__init__(self, settings)
        pm.CppBase.__init__(self,
                            module_path=m_path,
                            module_name='Controller',
                            binding_class_name="binding_Controller",
                            # If pybind is not global installed, use the uncomment the following line and set $(VENV) 
                            # as also as $(PYVERS) in the target_include_directories string.
                            # additional_lib={'pybind11': addLib}
                            )

        self.last_time = 0
        self.last_u = 0

        self.pid = self.get_class_from_module().PIDController(self._settings["Kp"],
                                                              self._settings["Ti"],
                                                              self._settings["Td"],
                                                              self._settings["output_limits"][0],
                                                              self._settings["output_limits"][1],
                                                              self._settings['dt [s]'])

    def _control(self,
                 time,
                 trajectory_values=None,
                 feedforward_values=None,
                 input_values=None,
                 **kwargs):
        # step size
        dt = time - self.last_time

        # input abbreviations
        x = np.zeros((len(self._settings["input_state"]),))
        for idx, state in enumerate(self._settings["input_state"]):
            x[idx] = input_values[int(state)]

        if np.isclose(dt, self._settings['dt [s]']):
            # save last control time
            self.last_time = time

            yd = trajectory_values

            u = self.pid.compute(x, yd)
        else:
            u = self.last_u

        self.last_u = u

        return u


class CppStateController(pm.CppBase, pm.Controller):
    """
    State Controller implemented in cpp with pybind11
    """
    public_settings = OrderedDict([
        ("poles", st.poles_state),
        ("dt [s]", 0.1),
        ("output_limits", st.limits_ctrl),
        ("input_state", st.input_ctrl),
        ("tick divider", 1),
    ])

    def __init__(self, settings):
        # add specific "private" settings
        settings.update(input_order=0)
        settings.update(output_dim=1)
        settings.update(input_type="system_state")

        proj_dir = os.path.abspath(os.path.dirname(__file__))
        m_path = os.sep.join([proj_dir, "src"])

        addLib = "add_library(pybind11 INTERFACE)\n"
        addLib += "target_include_directories(pybind11 INTERFACE $(VENV)/lib/$(PYVERS)/site-packages/pybind11/include/)\n"

        pm.Controller.__init__(self, settings)
        pm.CppBase.__init__(self,
                            module_path=m_path,
                            module_name='Controller',
                            binding_class_name="binding_Controller",
                            # If pybind is not global installed, use the uncomment the following line and set $(VENV) 
                            # as also as $(PYVERS) in the target_include_directories string.
                            # additional_lib={'pybind11': addLib}
                            )
        self.last_time = 0
        self.last_u = 0

        # pole placement of linearized state feedback
        x10, x20, uA0, A, B, C = self.calc_lin_sys()
        self.x0 = np.array([x10, x20])
        self._K = pm.controltools.place_siso(A, B, self._settings["poles"])[0]
        self._V = pm.controltools.calc_prefilter(A, B, C, self._K)[0][0]

        self.state = self.get_class_from_module().StateController(self._K,
                                                                  self._V,
                                                                  np.array([x10, x20, uA0]),
                                                                  self._settings["output_limits"][0],
                                                                  self._settings["output_limits"][1])

    def calc_lin_sys(self):
        x20 = 0.15

        K, AT, Aout1, Aout2, g, x1, x2, uA = sp.symbols('K A_T A_out1 A_out2 g x_1 x_2 u_A')

        # equilibrium
        dx1 = K / AT * uA - Aout1 / AT * sp.sqrt(2 * g * (x1 - x2))
        dx2 = Aout1 / AT * sp.sqrt(2 * g * (x1 - x2)) - Aout2 / AT * sp.sqrt(2 * g * x2)

        x0 = solve([dx1, dx2], [x1, uA], exclude=[K, AT, Aout1, Aout1, x2])[0]

        x10 = x0[0].subs([(K, st.K), (AT, st.AT), (Aout1, st.AS1), (Aout2, st.AS2), (g, st.g), (x2, x20)])
        uA0 = x0[1].subs([(K, st.K), (AT, st.AT), (Aout1, st.AS1), (Aout2, st.AS2), (g, st.g), (x2, x20)])

        subs_list = [(K, st.K), (AT, st.AT), (Aout1, st.AS1), (Aout2, st.AS2), (g, st.g), (x2, x20), (x1, x10),
                    (uA, uA0)]
        A = sp.Matrix([[dx1], [dx2]]).jacobian(sp.Matrix([[x1], [x2]])).subs(subs_list)
        B = sp.Matrix([[dx1], [dx2]]).jacobian(sp.Matrix([uA])).subs(subs_list)
        C = sp.Matrix([0, 1]).T

        return np.array(x10.evalf()).astype(np.float64), \
               x20, \
               np.array(uA0.evalf()).astype(np.float64), \
               np.array(A.evalf()).astype(np.float64), \
               np.array(B.evalf()).astype(np.float64), \
               np.array(C.evalf()).astype(np.float64)

    def _control(self,
                 time,
                 trajectory_values=None,
                 feedforward_values=None,
                 input_values=None,
                 **kwargs):
        # step size
        dt = time - self.last_time

        if np.isclose(dt, self._settings['dt [s]']):
            # save last control time
            self.last_time = time

            yd = trajectory_values

            u = self.state.compute(np.array(input_values), yd[0])
        else:
            u = self.last_u

        self.last_u = u

        return u


pm.register_simulation_module(pm.Controller, OpenLoop)
pm.register_simulation_module(pm.Controller, CppPIDController)
pm.register_simulation_module(pm.Controller, CppStateController)



#!/usr/bin/python2
# -*- coding: utf-8 -*-

import numpy as np
import os

import sys
sys.path.append(os.path.pardir)

import settings as st
#-------------------------------------------------------------
# settings
#-------------------------------------------------------------
end_time = 15
yd = [-4, 0] #[m]
dt = 5 #[s]

controllerList = ['FController', 'GController', 'JController',\
                'LSSController', 'PIFeedbackController', 'all']

#-------------------------------------------------------------
# init
#-------------------------------------------------------------
print('\n ### Regimefile Generator Eval C2 ### \n')

#print 'Choose parameter to test: '
#for idx, param in enumerate(st.paramVariationDictC1.keys()):
    #print '\t',idx,' - ', param

#paramIdx = -1
#while paramIdx not in range(0, len(st.paramVariationDictC1.keys())):
    #paramIdx = input()

paramIdx = 0

# extract parameter values
parameter = list(st.paramVariationDictC2.keys())[paramIdx]
lower_bound = st.paramVariationDictC2[parameter]['lower_bound']
upper_bound = st.paramVariationDictC2[parameter]['upper_bound']
step_size = st.paramVariationDictC2[parameter]['step_size']

#simLimits = np.arange(lower_bound, upper_bound + step_size, step_size)
#simLimits = [[-1*round(x,1), round(x,1)] for x in simLimits]
simLimits = [[-1*round(lower_bound, 1), round(lower_bound, 1)]]

print('Choose controller: ')
for idx, controller in enumerate(controllerList):
    print('\t', idx, ' - ', controller)

controlIdx = -1
while controlIdx not in list(range(0, len(controllerList))):
    controlIdx = eval(input())

if controlIdx == len(controllerList)-1:
    simCases = controllerList[:-1]
else:
    simCases = [controllerList[controlIdx]]

#-------------------------------------------------------------
# functions
#-------------------------------------------------------------

#prefined strings
def writeRegime(cName, pole, param, paramVal, appendix):
    return '- name: C2_' + cName + '_poles(' + str(pole) + ')'\
                    + '_' + parameter +'(' + str(paramVal) + ')'\
                    + appendix + '\n'\
                    + '  clear previous: !!python/bool False \n\n'

def writeModel(param, paramVal):
    return '  model:' + '\n'\
            + '   type: BallBeamModel' + '\n'\
            + '   ' + parameter + ': ' + str(paramVal) + '\n\n'

def writeController(cName, pole, multiplicator):
    tmp = '  controller: ' + '\n'\
            + '   type: '  + controller + '\n'\
            + '   poles: ' + str([pole]*multiplicator) + '\n'

    if (controller == 'PIFeedbackController' or controller == 'LSSController'):
        tmp += '   r0: ' + str(yd[1]) + '\n'
    tmp += '\n'
    return tmp

def writeSensor(senName, paramVal):
    return '  sensor:' + '\n'\
            + '   type: ' + senName + '\n'\
            + '   delay: ' + str(paramVal) + '\n'\
            + '   output: ' + '[0,1,2,3]' + '\n\n'

def writeDisturbance(disName, paramVal):
    if paramVal == 0:
        return '  disturbance:' + '\n'\
            + '   type: None\n\n'

    return '  disturbance:' + '\n'\
            + '   type: ' + disName + '\n'\
            + '   mean value: ' + '0' + '\n\n'\

def writeLimiter(limName, limits):
    return '  limiter:' + '\n'\
            + '   type: ' + limName + '\n'\
            + '   limits: ' + '[' + str(limits[0])+', '+str(limits[1])+']\n\n'\

def saveOutput(output, cName, poles, parameter, limits):
    '''
        wrapper to save regime file
    '''
    dirPath = os.path.join(os.path.pardir, os.path.pardir, 'regimes', 'generated')

    if not os.path.isdir(dirPath):
        os.makedirs(dirPath)

    fileName = 'C2_' + cName + '_poles' + str(poles) \
                        + '_' + str(parameter)\
                        + '(' + str(simLimits[0]) + ',' + str(simLimits[-1])\
                        + ').sreg'

    filePath = os.path.join(dirPath, fileName)
    with open(filePath, 'w') as f:
        f.write(output)

#-------------------------------------------------------------
# preamble
#-------------------------------------------------------------
preamble = '- name: C2-simulation-setup\n'\
+'  clear previous: !!python/bool True\n'\
+'\n'\
+'  model:\n'\
+'   type: BallBeamModel\n'\
+'\n'\
+'  solver:\n'\
+'   type: VODESolver\n'\
+'   initial state: ['+str(yd[0])+', 0, 0, 0]\n'\
+'   end time: '+str(end_time)+'\n'\
+'\n'\
+'  trajectory:\n'\
+'   type: SmoothTransitionTrajectory\n'\
+'   Positions: [' + str(yd[0]) + ', ' + str(yd[1]) + ']\n'\
+'   start time: 0\n'\
+'   delta t: ' + str(dt) + '\n'\
+'\n'

#-------------------------------------------------------------
# main
#-------------------------------------------------------------
collection = ''
for controller in simCases:
    lines = preamble

    # how many poles?
    if controller == 'PIFeedbackController':
        multiplicator = 5
    else:
        multiplicator = 4

    #set correct poles
    poles = st.smoothPoles[controller]

    #search limits
    for val in simLimits:
        #one without limitation
        lines += writeRegime(controller, poles, parameter, val, '_unlimited')
        lines += writeController(controller, poles, multiplicator)
        lines += '\n\n'
        #one with limitation
        lines += writeRegime(controller, poles, parameter, val, '_limited')
        lines += writeController(controller, poles, multiplicator)
        lines += writeLimiter('AmplitudeLimiter', val)
        lines += '\n\n'

        saveOutput(lines, controller, poles, parameter, simLimits)

    collection += lines

if len(simCases) > 1:
    saveOutput(collection, 'all', '_', parameter, simLimits)

print('done.')



# -*- coding: utf-8 -*-

import logging
import os
import sys
import subprocess
from pathlib import Path
import importlib.util

try:
    import pybind11
except ModuleNotFoundError:
    logging.warning("Package 'pybind11' could not be loaded, binding modules "
                    "will not be available")

__all__ = ["CppBase"]


BUILD_DIR = "_build"
LIB_DIR = "_lib"
CMAKE_LISTS = "CMakeLists.txt"


class BindingException(Exception):
    """
    Exception to be raised if the cpp binding raises.
    """
    pass


class CppBase:
    """
    Mix-in class for modules written in C++.

    This class uses pybind and cmake to automatically compile and link the
    provided sources and load them as an python module.

    Args:
        :module_path: Path to directory that contains the sources.
        :module_name: Name of the cpp class to use
        :binding_class_name: Name of the file including the binding definition
        :additional_lib: dict with key 'lib name' and additional lines for the
          CMakeLists

    Warn:
        The `module_name` will be used to generate the cmake configuration an,
        thus, expects ${module_name}.cpp and ${module_name}.h files.
        *Every other file* is ignored by this routine so far.

    """
    def __init__(self,
                 module_path=None,
                 module_name=None,
                 binding_class_name=None,
                 additional_lib=None):
        self._logger = logging.getLogger(self.__class__.__name__)

        # adapt to os-specific extensions
        if os.name == 'nt':
            self.sfx = '.pyd'
        else:
            self.sfx = '.so'

        if module_name is None:
            self._logger.error("Instantiation of binding class without"
                               " module_name is not allowed!")
            raise BindingException("Instantiation of binding class without"
                                   " module_name is not allowed!")
        self.module_name = module_name

        if module_path is None:
            self._logger.error("Instantiation of binding class without"
                               " module_path is not allowed!")
            raise BindingException("Instantiation of binding class without"
                                   " module_path is not allowed!")

        self.module_path = Path(module_path)
        self.module_stem = self.module_path / self.module_name
        self.module_inc_path = self.module_stem.with_suffix(".h")
        self.module_src_path = self.module_stem.with_suffix(".cpp")
        self.cmake_lists_path = self.module_path / CMAKE_LISTS
        self.module_build_path = self.module_path / BUILD_DIR
        self.module_lib_path = self.module_path / LIB_DIR

        self.additional_lib = None
        if additional_lib is not None:
            assert (isinstance(additional_lib, dict))
            self.additional_lib = additional_lib

        if binding_class_name is None:
            self._logger.error("Instantiation of binding class without"
                               " binding_class_name is not allowed!")
            raise BindingException("Instantiation of binding class without"
                                   " binding_class_name is not allowed!")
        self.binding_class_name = binding_class_name

        if self.create_binding_config():
            self.build_binding()
            self.install_binding()

    def create_binding_config(self):
        # check if folder exists
        if not self.module_path.is_dir():
            self._logger.error("Config folder '{}' does not exist."
                               "".format(self.module_path))
            return False

        if not self.module_src_path.is_file():
            self._logger.error("CPP binding '{}' could not be found in the "
                               "given module path '{}'."
                               "".format(self.module_src_path,
                                         self.module_path))
            return False

        if not self.module_inc_path.is_file():
            self._logger.error("Header file '{}' could not be found in the "
                               "given module path '{}'."
                               "".format(self.module_inc_path,
                                         self.module_path))
            return False

        if not self.cmake_lists_path.is_file():
            self._logger.warning("CMakeLists.txt not found in module path.")
            self._logger.info("Generating new CMake config.")
            self.create_cmake_lists()

        config_changed = self.update_binding_config()
        if config_changed:
            self.build_config()

        return True

    def create_cmake_lists(self):
        """
        Create the stub of a `CMakeLists.txt` .

        Returns:
        """
        c_make_lists = "cmake_minimum_required(VERSION 3.15)\n"
        c_make_lists += "project(Bindings)\n\n"

        c_make_lists += "set (Python_FIND_VIRTUALENV STANDARD)\n"
        py_ver = sys.version_info
        c_make_lists += "find_package(Python {}.{} EXACT REQUIRED " \
                        "COMPONENTS Interpreter Development)\n\n" \
                        "".format(py_ver.major, py_ver.minor)

        c_make_lists += "set( CMAKE_CXX_STANDARD 11 )\n\n"
        c_make_lists += "set( CMAKE_RUNTIME_OUTPUT_DIRECTORY . )\n"
        c_make_lists += "set( CMAKE_LIBRARY_OUTPUT_DIRECTORY . )\n"
        c_make_lists += "set( CMAKE_ARCHIVE_OUTPUT_DIRECTORY . )\n\n"

        c_make_lists += "foreach( OUTPUTCONFIG ${CMAKE_CONFIGURATION_TYPES} )\n"
        c_make_lists += "\tstring( TOUPPER ${OUTPUTCONFIG} OUTPUTCONFIG )\n"
        c_make_lists += "\tset( CMAKE_RUNTIME_OUTPUT_DIRECTORY_${OUTPUTCONFIG} . )\n"
        c_make_lists += "\tset( CMAKE_LIBRARY_OUTPUT_DIRECTORY_${OUTPUTCONFIG} . )\n"
        c_make_lists += "\tset( CMAKE_ARCHIVE_OUTPUT_DIRECTORY_${OUTPUTCONFIG} . )\n"
        c_make_lists += "endforeach( OUTPUTCONFIG CMAKE_CONFIGURATION_TYPES )\n\n"

        c_make_lists += "include_directories(${Python_INCLUDE_DIRS})\n"
        pybind_headers = Path(pybind11.get_include())
        c_make_lists += "include_directories({})\n".format(pybind_headers.as_posix())

        if self.additional_lib:
            c_make_lists += "\n\n"
            for _, value in self.additional_lib.items():
                c_make_lists += value
            c_make_lists += "\n\n"

        with open(self.cmake_lists_path, "w") as f:
            f.write(c_make_lists)

    def update_binding_config(self):
        """
        Add the module config to the cmake lists.

        Returns:
            bool: True if build config has been changed and cmake has to be
            rerun.
        """
        config_line = "add_library({} SHARED {} {})\n".format(
            self.module_name,
            self.module_src_path.as_posix(),
            self.binding_class_name + '.cpp'
        )
        config_line += "set_target_properties({} PROPERTIES PREFIX \"\" OUTPUT_NAME \"{}\" SUFFIX \"{}\")\n".format(
            self.module_name,
            self.module_name,
            self.sfx
        )
        
        config_target_line = lambda lib : "\ttarget_link_libraries({} ${{{}}}".format(
            self.module_name, lib, 
        )
        
        config_target_libs = ""
        if self.additional_lib:
            for key, _ in self.additional_lib.items():
                config_target_libs += " {}".format(key)
        config_target_libs += ")\n"

        
        config_line += "if(WIN32 AND MSVC)\n"
        config_line += config_target_line("Python_LIBRARY_RELEASE")
        config_line += config_target_libs
        config_line += "else()\n"
        config_line += config_target_line("PYTHON_LIBRARIES")
        config_line += config_target_libs
        config_line += "endif()\n"
        
        config_line += "install(FILES {}/{} DESTINATION {})".format(
            BUILD_DIR,
            self.module_name + self.sfx,
            self.module_lib_path.as_posix()
        )
        with open(self.cmake_lists_path, "r") as f:
            if config_line in f.read():
                return False

        self._logger.info("Appending build info for '{}'".format(self.module_name))
        with open(self.cmake_lists_path, "a") as f:
            f.write("\n")
            f.write(config_line)

        return True

    def build_config(self):
        # generate config
        if os.name == 'nt' and 'GCC' not in sys.version:
            cmd = ['cmake', '-A', 'x64', '-S', '.', '-B', BUILD_DIR]
        else:
            cmd = ['cmake',  # '-DCMAKE_BUILD_TYPE=Debug',
                   '-S', '.', '-B', BUILD_DIR]
        result = subprocess.run(cmd, cwd=self.module_path)

        if result.returncode != 0:
            self._logger.error("Generation of binding config failed.")
            raise BindingException("Generation of binding config failed.")

    def build_binding(self):
        # build
        if os.name == 'nt' and 'GCC' not in sys.version:
            cmd = ['cmake', '--build', BUILD_DIR, '--config', 'Release', '--target', 'INSTALL']
        else:
            cmd = ['cmake', '--build', BUILD_DIR]
        result = subprocess.run(cmd, cwd=self.module_path)

        if result.returncode != 0:
            self._logger.error("Build failed!")
            raise BindingException("Build failed!")

    def install_binding(self):
        # generate config
        if os.name == 'nt' and 'GCC' not in sys.version:
            return
        else:
            cmd = ['cmake', '--install', BUILD_DIR]
        result = subprocess.run(cmd, cwd=self.module_path)

        if result.returncode != 0:
            self._logger.error("Installation of bindings failed.")
            raise BindingException("Installation of bindings failed.")

    def get_class_from_module(self):
        try:
            spec = importlib.util.spec_from_file_location(self.module_name,
                                                          (self.module_lib_path / self.module_name).with_suffix(self.sfx))
            module = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(module)
            return module
        except ImportError as e:
            self._logger.error("Cannot load module: {}".format(e))
            raise e



# -*- coding: utf-8 -*-

import numpy as np
from collections import OrderedDict

import pymoskito as pm

from . import settings as st


class BallInTubeFeedforward(pm.Feedforward):
    """
    Calculate flatness based feedforward.

    The flat output is given by the ball position (x3)
    x1_flat and x2_flat are the system states expressed in flat coordinates.
    """
    public_settings = OrderedDict([("tick divider", 1)])

    def __init__(self, settings):
        settings.update(input_order=4)
        settings.update(output_dim=1)
        pm.Feedforward.__init__(self, settings)

    def _feedforward(self, time, trajectory_values):
        yd = trajectory_values[0]
        x1_flat = (np.sqrt((yd[2] + st.g)*st.m*st.A_Sp**2/st.k_L)
                   + st.A_B*yd[1])/st.k_V
        x2_flat = (yd[3]*st.m*st.A_Sp**2/(2*st.k_V*st.k_L*(st.k_V*x1_flat
                                                           - st.A_B*yd[1]))
                   + st.A_B*yd[2]/st.k_V)

        # feed forward control
        u = (st.T**2*(yd[4]*st.m*st.A_Sp**2 - 2*st.k_L*(st.k_V*x2_flat
                                                        - st.A_B*yd[2])**2)
             / (2*st.k_s*st.k_V*st.k_L*(st.k_V*x1_flat - st.A_B*yd[1]))
             + (st.T**2*st.A_B*yd[3])/(st.k_s*st.k_V)
             + (2*st.d*st.T*x2_flat)/st.k_s
             + x1_flat/st.k_s)*(255/st.Vcc)

        return np.array([[u]], dtype=float)

pm.register_simulation_module(pm.Feedforward, BallInTubeFeedforward)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
test_processing
----------------------------------

Tests for `processing` module.
"""

import unittest
import os

from PyQt5.QtWidgets import QApplication

from pymoskito.processing_gui import PostProcessor


@unittest.skip  # TODO find a way to generate test data
class TestProcessing(unittest.TestCase):

    def setUp(self):
        self.app = QApplication([])
        self.post = PostProcessor()

        # load some test data sets
        path = os.path.join("data", "step_response.pmr")
        self.post._load_result_file(path)

    def test_postprocessor(self):
        self.post.run_processor("StepResponse", processor_type="post")
        self.assertTrue(os.path.isfile(os.path.join("results",
                                                    "processing",
                                                    "StepResponse",
                                                    "test-nonlinear.pof")))

    def tearDown(self):
        del self.post

if __name__ == '__main__':
    unittest.main()



from . import model
from . import controller
from . import visualizer_mpl
# from . import visualizer_vtk

import pymoskito as pm

# register model
pm.register_simulation_module(pm.Model, model.PendulumModel)

# register controller
pm.register_simulation_module(pm.Controller, controller.BasicController)

# register visualizer
pm.register_visualizer(visualizer_mpl.MplPendulumVisualizer)
# pm.register_visualizer(visualizer_vtk.VtkPendulumVisualizer)



import pymoskito as pm


class PolesITAEMetaProcessor(pm.XYMetaProcessor):
    """
    Metaprocessor to evaluate the ITAE error for different pole placements of
    different controllers.

    Run the `StepResponse` postprocessor with the results from the `pp_regimes`
    file the create the required input data.
    """

    def __init__(self):
        sort_key = ["modules", "Controller", "type"]
        x_path = ["modules", "Controller", "poles"]
        y_path = ["metrics", "L1NormITAE"]

        pm.XYMetaProcessor.__init__(self, x_path, y_path, sort_key, x_idx=0,
                                    title="ITAE Error over Pole Placement",
                                    x_label="Pole 0", y_label="ITAE Error",
                                    # line_style="bar"
        )


class MassMetricMetaProcessor(pm.XYMetaProcessor):

    def __init__(self):
        sort_key = ["modules", "Controller", "type"]
        x_path = ["modules", "Model", "M"]
        y_path = ["metrics", "L1NormITAE"]

        pm.XYMetaProcessor.__init__(self, x_path, y_path, sort_key)


pm.register_processing_module(pm.MetaProcessingModule, PolesITAEMetaProcessor)
pm.register_processing_module(pm.MetaProcessingModule, MassMetricMetaProcessor)
